// 
// Menu for configuring X terminals.
//

#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <unistd.h>
#include <regex.h>
#include <misc.h>
#include <fviews.h>
#include <dialog.h>
#include <translat.h>
#include <private_msg.h>


// Store all the user supplied configuration parameters for an X terminal.

class TERMINAL_CONFIG {
  public:
  SSTRING termname;         // e.g. station12
  SSTRING keyboard;         // list /usr/lib/kbd/keymaps/i386/*/*.keymap.gz
  char emulate_three;       // yes (y), no(n)
  char local_printer;       // yes (y), no (n)
  char indirect;			// Support indirect login instead of -query
  SSTRING wanted_resolution;// e.g. 800x600
  SSTRING bits_per_pixel;   // e.g. 8
  SSTRING monitor_hsync;    // e.g. 30-64
  SSTRING monitor_vsync;    // e.g. 50-100
  SSTRING swap_partition;   // e.g. /dev/hda1
  SSTRING swap_partition_type; // none (whole native swap), msdos (swap file)
  SSTRING boot_partition;   // e.g. /dev/hda1
  SSTRING boot_type;        // loadlin, lilo
  SSTRING defprinter;       // printer name, e.g. lp
  SSTRING term_location;    // room number? e.g. C-631
  SSTRING term_groups;      // grade5, class3
  SSTRING application_server; // e.g. defserver
  SSTRING font_server;      // e.g. defserver
  // These parameters are not user supplied. They could be shown for 
  // information purposes.
  //  SSTRING mouse_dev;        // e.g. /dev/psaux
  //  SSTRING sound_card;       // e.g. CS4232
  //  SSTRING xserver;          // e.g. SVGA
};


// All the configuration files accessed by the program

static char *NEW_file = "/etc/sysconfig/NEW";

static CONFIG_FILE keyboard_file("/etc/sysconfig/keyboard",
   help_nil,
   CONFIGF_MANAGED | CONFIGF_OPTIONAL,"root","root",0644);

static CONFIG_FILE mouse_file("/etc/sysconfig/mouse",
   help_nil,
   CONFIGF_MANAGED | CONFIGF_OPTIONAL,"root","root",0644);

static CONFIG_FILE swap_file("/etc/sysconfig/swap",
   help_nil,
   CONFIGF_MANAGED | CONFIGF_OPTIONAL,"root","root",0644);

static CONFIG_FILE xterminal_file("/etc/sysconfig/xterminal",
   help_nil,
   CONFIGF_MANAGED | CONFIGF_OPTIONAL,"root","root",0644);

static CONFIG_FILE termname_file("/etc/sysconfig/TERMNAME",
   help_nil,
   CONFIGF_MANAGED | CONFIGF_OPTIONAL,"root","root",0644);

static CONFIG_FILE defprinter_file("/etc/sysconfig/DEFPRINTER",
   help_nil,
   CONFIGF_MANAGED | CONFIGF_OPTIONAL,"root","root",0644);

static CONFIG_FILE printer_file("/etc/sysconfig/printserv",
   help_nil,
   CONFIGF_MANAGED | CONFIGF_OPTIONAL,"root","root",0644);

static CONFIG_FILE X_file("/etc/X11/XF86Config",
   help_nil,
   CONFIGF_MANAGED | CONFIGF_OPTIONAL,"root","root",0644);


// Read the value of a symbolic link. Currently not used.
/*
static void read_link(char *path, SSTRING &link_value) {
  char buffer[PATH_MAX];

  int len = readlink(path,buffer,sizeof(buffer)-1);
  if (len != -1){
    buffer[len] = '\0';
    link_value.setfrom(buffer);
  }
  else link_value.setfrom("");
}
*/

// Read a single line file. Return an empty string if the file is not there.

static void read_file(CONFIG_FILE &file, SSTRING &string) {
  VIEWITEM *item;
  VIEWITEMS items;

  int n;
  const char *pt;

  string.setfrom("");
  items.read(file);
  n = items.getnb();
  if(n > 0) {
    item = items.getitem(0);
    pt = item->line.get();
    if(pt != NULL && strlen(pt) > 0) string.setfrom(pt);
  }
}


// Write a single-line file.

static void write_file(CONFIG_FILE &file, SSTRING &string) {
  VIEWITEM item(string.get());
  VIEWITEMS items;

  items.neverdelete();
  items.add(&item);
  items.write(file,(PRIVILEGE *)NULL);
}


// Simple data structures to store the partition list.

class PARTITION: ARRAY_OBJ {
public:
  SSTRING device;
  char bootable;
  SSTRING type;
};


class PARTITIONS: public ARRAY{
public:
  PARTITION *getitem (int no) const;
  void add (PARTITION *pt);
};


PARTITION *PARTITIONS::getitem(int no) const
{
  return (PARTITION*)ARRAY::getitem(no);
}


void PARTITIONS::add(PARTITION *pt)
{
  ARRAY::add((ARRAY_OBJ *)pt);
}


// List the partitions. The devices with partitions are obtained from
// /proc/partitions and used as arguments to fdisk.

static void list_partitions(PARTITIONS &partitions) {
  char *command = "fdisk -l `cat /proc/partitions | grep '\\(hd[a-z][0-9][0-9]*\\)\\|\\(sd[a-z][0-9][0-9]*\\)' | sed -e 's+.*\\(hd[a-z]\\).*+/dev/\\1+' -e 's+.*\\(sd[a-z]\\).*+/dev/\\1+' | uniq`";

  char buffer[100], word[100];
  char *pt;
  SSTRING string;
  PARTITION *part;

  partitions.remove_all();

  // Obtain the partitions, one per line, from the piped command.
  // Each line contains:
  //
  // /dev/hda5             1       509   4088479+  83  Linux

  FILE *pipe = popen (command, "r");
  if(pipe != NULL) {
    while(fgets(buffer,sizeof(buffer) - 1,pipe) != NULL) {
      if(buffer[0] != '/') continue;
      part = new PARTITION;
      pt = str_copyword(word,buffer);
      part->device.setfrom(word);
      pt = str_skip(pt);
      if(pt[0] == '*') {
        part->bootable = 1;
        pt = str_skipword(pt);
        pt = str_skip(pt);
      }
      else part->bootable = 0;

      // skip the start, end, and size in blocks
      pt = str_skipword(pt);
      pt = str_skip(pt);
      pt = str_skipword(pt);
      pt = str_skip(pt);
      pt = str_skipword(pt);
      pt = str_skip(pt);

      // Here we have the partition type
      pt = str_copyword(word,pt);
      if(strcmp(word,"83") == 0) part->type.setfrom("linux");
      else if(strcmp(word,"82") == 0) part->type.setfrom("swap");
      else if(strcmp(word,"4") == 0 || strcmp(word,"6") == 0) {
        part->type.setfrom("msdos");
      }
      else {
        // Unknown type, probably an extension partition
        free(part);
        continue;
      }
      partitions.add(part);
    }
  }
  pclose(pipe);
}


// Obtain the list of keymaps from a recursive list /usr/lib/kbd/keymaps/i386

static void list_keymaps(SSTRINGS &strings) {
  SSTRING command;
  const char *dir = "/usr/lib/kbd/keymaps";
  if (!file_exist (dir)) dir = "/lib/kbd/keymaps";
  command.setfromf ("find %s/i386 -name '*.kmap.*'"
		" -printf '%%f\\n' | sed -e 's+\\.kmap\\..*++'"
		,dir);
  char buffer[100];

  strings.remove_all();

  FILE *pipe = popen (command.get(), "r");
  if(pipe != NULL) {
    while(0 <= fscanf(pipe,"%99s",buffer)) {
      strings.add(new SSTRING(buffer));
    }
    pclose(pipe);
    strings.sort();
  }
}


// Translate from console keymap name to XFree keymap name

static void keymap_translate(SSTRING &keymap, SSTRING &xkeymap) {
  const char *k = keymap.get();

  if(strlen(k) < 2) xkeymap.setfrom("us");
  else if(strcmp(k,"cf") == 0) xkeymap.setfrom("ca_enhanced");
  else if(strcmp(k,"uk") == 0) xkeymap.setfrom("gb");
  else xkeymap.setfrom(k,2);
}


// Read the terminal configuration from various files into the 
// TERMINAL_CONFIG structure.

static void load_terminal_config(TERMINAL_CONFIG &config) {
  VIEWITEMS items;
  char buffer[1000];

  read_file(termname_file,config.termname);
  read_file(defprinter_file,config.defprinter);

  items.read(keyboard_file);
  config.keyboard.setfrom(items.locateval("KEYTABLE",buffer));
  items.remove_all();

  items.read(mouse_file);
  config.emulate_three = items.locatebval("XEMU3");
  items.remove_all();

  items.read(xterminal_file);
  config.term_location.setfrom(items.locateval("TERMLOCATION",buffer));
  config.term_groups.setfrom(items.locateval("TERMGROUPS",buffer));
  config.application_server.setfrom(items.locateval("APPSERVER",buffer));
  config.font_server.setfrom(items.locateval("FONTSERVER",buffer));
  config.wanted_resolution.setfrom(items.locateval("RESOLUTION",buffer));
  config.bits_per_pixel.setfrom(items.locateval("BITSPERPIXEL",buffer));
  config.monitor_hsync.setfrom(items.locateval("HSYNC",buffer));
  config.monitor_vsync.setfrom(items.locateval("VSYNC",buffer));
  config.boot_partition.setfrom(items.locateval("BOOTDEV",buffer));
  config.boot_type.setfrom(items.locateval("BOOT",buffer));
  config.indirect = items.locatebval("INDIRECT");
  items.remove_all();

  items.read(swap_file);
  config.swap_partition.setfrom(items.locateval("SWAPDEV",buffer));
  config.swap_partition_type.setfrom(items.locateval("SWAPFS",buffer));
  items.remove_all();
  
  items.read(printer_file);
  config.local_printer = items.locatebval("PRINTSERV");
  items.remove_all();
}


// Write to the various configuration files the content of the TERMINAL_CONFIG
// data structure.

static void save_terminal_config(TERMINAL_CONFIG &config) {
  VIEWITEMS items;
  int ret;

  write_file(termname_file,config.termname);
  write_file(defprinter_file,config.defprinter);

  items.read(keyboard_file);
  items.update("KEYTABLE",config.keyboard);
  ret = items.write(keyboard_file,(PRIVILEGE *)NULL);
  items.remove_all();
  //xconf_notice("Write keyboard file ret %d",ret);

  items.read(mouse_file);
  items.updatebval("XEMU3",config.emulate_three);
  ret = items.write(mouse_file,(PRIVILEGE *)NULL);
  items.remove_all();

  items.read(xterminal_file);
  items.update("TERMLOCATION",config.term_location);
  items.update("TERMGROUPS",config.term_groups);
  items.update("APPSERVER",config.application_server);
  items.update("FONTSERVER",config.font_server);
  items.update("RESOLUTION",config.wanted_resolution);
  items.update("BITSPERPIXEL",config.bits_per_pixel);
  items.update("HSYNC",config.monitor_hsync);
  items.update("VSYNC",config.monitor_vsync);
  items.update("BOOTDEV",config.boot_partition);
  items.update("BOOT",config.boot_type);
  items.updatebval("INDIRECT",config.indirect);
  items.write(xterminal_file,(PRIVILEGE *)NULL);
  items.remove_all();

  items.read(swap_file);
  items.update("SWAPDEV",config.swap_partition);
  items.update("SWAPFS",config.swap_partition_type);
  items.write(swap_file,(PRIVILEGE *)NULL);
  items.remove_all();
  
  items.read(printer_file);
  items.updatebval("PRINTSERV",config.local_printer);
  items.write(printer_file,(PRIVILEGE *)NULL);
  items.remove_all();
}


// Check if the first value on the line (second word) is at the specified
// value. Remove the quotes from the value on the line, if present, before
// checking.

static int check_value(const char *line, const char *value) {
  char word[101];
  int len;

  // Skip the key, and get the value. Remove surrounding quotes
  // before checking the value;
  line = str_skip(line);
  line = str_skipword(line);
  line = str_skip(line);
  if(*line == '"') line++;
  line = str_copyword(word,line,100);
  len = strlen(word);
  if(len > 0 && word[len - 1] == '"') word[len - 1] = 0;
  if(strcmp(word,value) == 0) return(1);
  else return 0;
}


// Check if item defined by key has the specified value

static int check_item(VIEWITEMS &items, const char *key, const char *value, 
    int start, int end) {
  VIEWITEM *item;

  // Locate the item to check
  item = items.locate(key,start,end);

  // No such item, fail.
  if(item == NULL) return -1;
  return(check_value(item->line.get(),value));
}


// Find the first item between start and end, with the specified key and value.

static int find_item(VIEWITEMS &items, const char *key, const char *value, 
    int start, int end) {
  VIEWITEM *item;

  for (int i = start; i < end; i++){
    item = items.locate(key,i,i);
    if(item != NULL && check_value(item->line.get(),value)) return i;
  }
  return -1;
}

static bool config_validhostorip (const SSTRING &s, bool maybe_empty)
{
	bool ret = false;
	if (s.is_empty()){
		ret = maybe_empty;
	}else if (ipnum_validip (s.get(),true)){
		ret = true;
	}else{
		const char *pt = s.get();
		bool last_was_pt = true;
		ret = true;
		while (*pt != '\0'){
			if (!isalpha(*pt)){
				// After a dot, we need an alpha character, and at the start
				if (!last_was_pt){
					ret = false;
					break;
				}else if (!isdigit(*pt) && *pt != '-'){
					ret = false;
					break;
				}
			}
			pt++;
		}
	}
	return ret;
}


// Fields validation:
//
//   proper names: termname, defprinter, application_server, font_server
//   enums: keyboard, wanted_resolution, bits_per_pixel, monitor_hsync,
//          monitor_vsync, swap_partition_type, swap_partition, 
//          boot_type, boot_partition
//   word: location
//   words: groups

static int validate_terminal_config(TERMINAL_CONFIG &config, int &nof)
{
  int ret;
  regex_t name, number, resolution, dev, freq;

  // For these fields use a simple regular expression based check

  ret = regcomp(&name,"^[a-zA-Z][a-zA-Z0-9]*$",REG_NOSUB | REG_EXTENDED);
  ret = regcomp(&number,"^[0-9][0-9]*$",REG_NOSUB | REG_EXTENDED);
  ret = regcomp(&resolution,"^[0-9][0-9]*x[0-9][0-9]*$",
      REG_NOSUB | REG_EXTENDED);
  ret = regcomp(&dev,"^/dev/[hs]d[a-z][0-9]{1,3}$",REG_NOSUB | REG_EXTENDED);
  ret = regcomp(&freq,"^[0-9][0-9]*(\\.[0-9]*)?(-[0-9][0-9]*(\\.[0-9]*)?)?$",
      REG_NOSUB | REG_EXTENDED);

  ret = 0;

  if(regexec(&name,config.termname.get(),0,NULL,0) == REG_NOMATCH) {
    xconf_error("invalid characters in terminal name");
    nof = 0;
    ret = -1;
  }

  else if(regexec(&resolution,config.wanted_resolution.get(),0,NULL,0) == 
      REG_NOMATCH) {
    xconf_error("invalid characters in screen resolution");
    nof = 4;
    ret = -1;
  }

  else if(regexec(&number,config.bits_per_pixel.get(),0,NULL,0) == 
      REG_NOMATCH) {
    xconf_error("invalid characters in bits per pixel");
    nof = 5;
    ret = -1;
  }

  else if(regexec(&freq,config.monitor_hsync.get(),0,NULL,0) == REG_NOMATCH) {
    xconf_error("invalid characters in horizontal frequency");
    nof = 6;
    ret = -1;
  }

  else if(regexec(&freq,config.monitor_vsync.get(),0,NULL,0) == REG_NOMATCH) {
    xconf_error("invalid characters in vertical frequency");
    nof = 7;
    ret = -1;
  }

  else if(strcmp(config.swap_partition_type.get(),"none") != 0 &&
      regexec(&dev,config.swap_partition.get(),0,NULL,0) == REG_NOMATCH){
    xconf_error("invalid characters in swap partition");
    nof = 8;
    ret = -1;
  }

  else if(strcmp(config.boot_type.get(),"none") != 0 && 
      regexec(&dev,config.boot_partition.get(),0,NULL,0) == REG_NOMATCH) {
    xconf_error("invalid characters in boot partition");
    nof = 10;
    ret = -1;
  }

  // optional, may be left blank.

  else if(strlen(config.defprinter.get()) > 0 && 
      regexec(&name,config.defprinter.get(),0,NULL,0) == REG_NOMATCH) {
    xconf_error("invalid characters in default printer");
    nof = 12;
    ret = -1;
  }

  // optional, may be left blank.

  else if(!config_validhostorip (config.application_server,true)){
    xconf_error("invalid characters in application server name");
    nof = 15;
    ret = -1;
  }

  // optional, may be left blank.

  else if(!config_validhostorip (config.font_server,true)){
    xconf_error("invalid characters in font server name");
    nof = 16;
    ret = -1;
  }

  regfree(&name);
  regfree(&number);
  regfree(&resolution);
  regfree(&dev);
  regfree(&freq);

  if(ret < 0) return ret;

  // Read XF86Config, find the relevant configuration parameters, and check
  // these against the user supplied values.

  SSTRING xkeymap;
  VIEWITEMS items;
  items.read (X_file);
  int n = items.getnb();
  int start = 0;
  char *msg;
  bool e3b;

  keymap_translate(config.keyboard,xkeymap);

  if(check_item(items,"XkbLayout",xkeymap.get(),0,n) != 1) {
    msg = "Keyboard definition";
    ret = -1; 
    nof = 2;
  }

  else if(((e3b = (items.locate("Emulate3Buttons") != NULL)) &&  
           !config.emulate_three) ||
          (!e3b && config.emulate_three)) {
    msg = "Emulate three button mouse";
    ret = -1; 
    nof = 3;
  }

  else if(check_item(items,"HorizSync",config.monitor_hsync.get(),0,n) != 1) {
    msg = "Screen horizontal frequency";
    ret = -1; 
    nof = 7;
  }

  else if(check_item(items,"VertRefresh",config.monitor_vsync.get(),0,n) != 1){
    msg = "Screen vertical frequency";
    ret = -1; 
    nof = 8;
  }

  // Not found (returns -1) is accepted because more difficult to check.
  // 0 means the wrong value and is a sure indication of problem.

  else if(check_item(items,"DefaultColorDepth",config.bits_per_pixel.get(),
      0,n) == 0) {
    msg = "Bits per pixel";
    ret = -1; 
    nof = 6;
  }

  // Check only the Modes for the sections with the Depth equal to the
  // DefaultDepth. The wanted resolution should appear first on the Modes
  // line.

  start = 0;
  while(ret >= 0 && (start = find_item(items,"Depth",
      config.bits_per_pixel.get(),start,n)) >= 0) {
    if(check_item(items,"Modes",config.wanted_resolution.get(),start,n) != 1) {
      msg = "Screen resolution";
      ret = -1; 
      nof = 5;
    }
  }

  // Some parameters differ in file XF86Config. Either Xconfigurator
  // was not run after some user supplied parameters were changed, or
  // Xconfigurator failed to provide the desired parameters. In the
  // first case the user return to the configuration screen, and in 
  // the second case the user cannot do much and should be allowed to proceed.
  // Instead of just failing the validation, prompt the user.

  char buffer[500];
  MENU_STATUS code;

  if(ret < 0) {
    sprintf(buffer,"%s differs in file \
/etc/X11/XF86Config, perhaps you did not reconfigure \
the graphics card and mouse, \
 \
Do you want to go back to the configuration screen?",msg);

    code = dialog_yesno("Graphics configuration", buffer, help_nil, true);

    // The user does not want to get back to configuration, accept the
    // values.
    if(code == MENU_NO) return 0; 
  }
  return ret;
}
     

// Configuration activation. The new configuration was saved, we simply
// have to check that the new configuration parameters are taken into
// account. Work to do for the different fields:
//
// Immediate fields (used when needed, thus nothing to update):
//   termname (update on server), defprinter, application_server, font_server
//
// Fields not yet used: location, groups
//
// The following fields are used by mouseconfig and Xconfigurator.
// The validation phase checks if the XF86Config file is in sync
// with these values.
//   keyboard, wanted_resolution, bits_per_pixel, monitor_hsync, monitor_vsync
//   emulate_three
//
// Call enable-swap which will check and activate as needed. We could call
// swapon -s to check here and avoid calling enable-swap.
//   swap_partition_type, swap_partition
//
// Call enable-boot which will check and activate the boot setup.
//   boot_type, boot_partition

static void activate_terminal_config(TERMINAL_CONFIG &config) {
  char buffer[1000];

  // We could call netconf_system to have timeouts
  // enable-printer??
  system("/bin/enable-swap >>/etc/sysconfig/config.log 2>&1");
  system("/bin/cdromconf >>/etc/sysconfig/config.log 2>&1");
  sprintf(buffer,"/bin/enable-boot %s %s >>/etc/sysconfig/config.log 2>&1",
            config.boot_type.get(),
	    config.boot_partition.get());

  system(buffer);
}


// Read the configuration, show the configuration parameters through the
// user interface, save the updated configuration.

int main (int argc, char *argv[])
{
  argc = dialog_parseuioptions (argc, argv);
  linuxconf_loadlibmsg("linuxconf");
  translat_checkmissing();
  dialog_clear();

  DIALOG dia;
  SSTRINGS keymaps;
  TERMINAL_CONFIG config;
  PARTITIONS partitions;
  bool new_config = file_exist(NEW_file);

  system("echo xtermconfig >/etc/sysconfig/config.log");
  load_terminal_config(config);
  list_keymaps(keymaps);
  int nb_keymaps = keymaps.getnb();
  list_partitions(partitions);
  int nb_partitions = partitions.getnb();

  dia.newf_str("Terminal name",config.termname);

  // Show all the available keymaps.

  FIELD_LIST *keyboard_list = dia.newf_list("keyboard",config.keyboard);
  for(int i = 0; i < nb_keymaps; i++) {
    keyboard_list->addopt(keymaps.getitem(i)->get(),"");
  }

  dia.newf_chk("Two buttons mouse",config.emulate_three,
      "(emulate third by pressing the two)");

  dia.newf_chk("Local printer(s)",config.local_printer,"attached");

  FIELD_LIST *resolution_list = dia.newf_list("Screen resolution",
      config.wanted_resolution);
  resolution_list->addopt("640x480","");
  resolution_list->addopt("800x600","");
  resolution_list->addopt("1024x768","");
  resolution_list->addopt("1280x1024","");
  resolution_list->addopt("1600x1200","");
  //resolution_list->addopt("1800X1440","");

  FIELD_LIST *bpp_list = dia.newf_list("Bits per pixel",
      config.bits_per_pixel);
  bpp_list->addopt("8","256 colors");
  bpp_list->addopt("16","65000 colors");
  bpp_list->addopt("24","true colors");
  bpp_list->addopt("32","true colors");

  FIELD_COMBO *hsync_list = dia.newf_list("Screen horizontal frequency range",
      config.monitor_hsync);
  hsync_list->addopt("31.5","KHz");
  hsync_list->addopt("31.5-35.1","KHz (640x480@60Hz)");
  hsync_list->addopt("31.5-37.9","KHz (800x600@56Hz)");
  hsync_list->addopt("31.5-48.5","KHz (800x600@72Hz)");
  hsync_list->addopt("31.5-57.0","KHz (1024x768@70Hz)");
  hsync_list->addopt("31.5-64.3","KHz (1280x1024@60Hz)");
  hsync_list->addopt("31.5-79.0","KHz (1280x1024@74Hz)");
  hsync_list->addopt("31.5-82.0","KHz (1280x1024@76Hz)");
  hsync_list->addopt("31.5-88.0","KHz (1600x1200@70Hz)");
  hsync_list->addopt("31.5-94.0","KHz (1600x1200@76Hz)");

  FIELD_COMBO *vsync_list = dia.newf_list("Screen vertical frequency range",
      config.monitor_vsync);
  vsync_list->addopt("50-70","Hz");
  vsync_list->addopt("50-90","Hz");
  vsync_list->addopt("50-100","Hz");
  vsync_list->addopt("40-150","Hz");

  // Show all local partitions.

  PARTITION *part; 
  FIELD_LIST *swap_list = dia.newf_list("Swap partition",
      config.swap_partition);
  for(int i = 0; i < nb_partitions; i++) {
    part = partitions.getitem(i);
    swap_list->addopt(part->device.get(),part->type.get());
  }
  swap_list->addopt("none","(no swap)");

  // The partition type could be inferred from the selected partition.
  // However we may eventually offer to reformat partitions and change
  // their type.

  FIELD_LIST *swap_type_list = dia.newf_list("Swap partition file system type",
      config.swap_partition_type);
  swap_type_list->addopt("none","(no swap)");
  swap_type_list->addopt("swap","(swap partition)");
  swap_type_list->addopt("msdos/ext2","(swap file in partition)");

  // Show all the available partitions.

  FIELD_LIST *boot_list = dia.newf_list("Boot partition",
      config.boot_partition);
  for(int i = 0; i < nb_partitions; i++) {
    part = partitions.getitem(i);
    if(part->bootable) boot_list->addopt(part->device.get(),part->type.get());
  }
  boot_list->addopt("none","(do not start Linux from local hard disk)");

  FIELD_LIST *boot_type_list = dia.newf_list("Boot type",
      config.boot_type);
  boot_type_list->addopt("none","(do not start Linux from local hard disk)");
  boot_type_list->addopt("loadlin","(start Linux from DOS on local disk)");
  boot_type_list->addopt("lilo","(native Linux loader on local disk)");

  dia.newf_str("Default printer",config.defprinter);
  dia.newf_str("Terminal location",config.term_location);
  dia.newf_str("Terminal groups",config.term_groups);
  dia.newf_str("Application server",config.application_server);
  dia.last_noempty();
  dia.newf_chk("Use the indirect",config.indirect,"protocol");
  dia.newf_str("Font server",config.font_server);
  dia.last_noempty();

  PRIVATE_MESSAGE config_mouse;
  dia.new_button("Configure mouse", "", config_mouse);
  PRIVATE_MESSAGE config_sound;
  dia.new_button("Configure sound", "", config_sound);
  PRIVATE_MESSAGE config_X;
  dia.new_button("Configure graphics card and screen", "",config_X);

  // This should be the most common operation. Make it a button besides
  // accept and cancel.

  dia.setbutinfo(MENU_USR1,"Autoconfig","Autoconfig");
  dia.setbutinfo(MENU_USR2,"Autoconfig as SVGA","Autoconfig");

  int nof = 0;
  SSTRING intro;
  if(new_config) {
    intro.append("New X terminal configuration, use autoconfigure\n");
  }
  else intro.append("Edit X terminal configuration\n");

  while(1) {
    MENU_STATUS code = dia.edit("X Terminal", intro.get(), help_nil,nof,
        MENUBUT_CANCEL | MENUBUT_ACCEPT | MENUBUT_USR1 | MENUBUT_USR2);

    if(code == MENU_CANCEL || code == MENU_ESCAPE){ break; }

    else if(code == MENU_ACCEPT) {

      // Do not accept if validation fails.
      if(validate_terminal_config(config,nof) == -1) {}
      else {
        save_terminal_config(config);
        activate_terminal_config(config);
        break;
      }
    }
    else if(code == MENU_MESSAGE) {
      if(dialog_testmessage(config_mouse)) {
        dialog_end();
        system("mouseconfig");
        dialog_restart();
      }
      if(dialog_testmessage(config_sound)) {
        dialog_end();
        system("sndconfig --confpath /etc/X11");
        dialog_restart();
      }
      if(dialog_testmessage(config_X)) {
        dialog_end();
        system("Xconfigurator");
        unlink(NEW_file);
        dialog_restart();
      }
    }
    else if(code == MENU_USR1 || code == MENU_USR2) {

      // Call the autoconfig script with all the needed parameters.
      dia.save();
      dialog_end();
      // We will autoprobe the card, which may hang. Saving is thus preferable
      save_terminal_config(config);

      char buffer[1000];
      int e3b = config.emulate_three;
      char *xfree_server = "PROBE";

      if(code == MENU_USR2) xfree_server = "SVGA";

      sprintf(buffer,
          "Xautoconfig \'%s\' \'%d\' \'%s\' \'%s\' \'%s\' \'%s\' \'%s\' >>/etc/sysconfig/config.log 2>&1",
          config.keyboard.get(), e3b, 
          config.wanted_resolution.get(), config.bits_per_pixel.get(),
          config.monitor_hsync.get(), config.monitor_vsync.get(), xfree_server);
      system(buffer);
      unlink(NEW_file);
      dialog_restart();
    }
  }
}

